"""
This module is about processing royalty statements.

It processes files in the request-response cycle, not in background workers.
Therefore, focus is on speed. Nothing is written to the database, and
SELECTs are optimised and performed in one batch.

"""

import csv
import os
from collections import defaultdict
from decimal import Decimal
from io import TextIOWrapper
from tempfile import NamedTemporaryFile

from django import forms
from django.conf import settings
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.core.files.uploadhandler import TemporaryFileUploadHandler
from django.http import FileResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.edit import FormView

from .models import SOCIETY_DICT, WorkAcknowledgement, Writer, WriterInWork


def get_id_sources():
    """
    Yield choices, fixed and societies.
    """
    yield None, 'Publisher Work ID:'
    yield settings.PUBLISHER_CODE, settings.PUBLISHER_NAME
    yield None, 'International Standard Codes:'
    yield 'ISWC', 'ISWC'
    yield 'ISRC', 'ISRC'
    yield None, 'Sender Work ID:'
    codes = WorkAcknowledgement.objects.order_by(
        'society_code').values_list('society_code', flat=True).distinct()
    codes = [(code, SOCIETY_DICT.get(code, '')) for code in codes]
    codes.sort(key=lambda code: code[1])
    for code in codes:
        yield code


def get_right_types():
    """
    Yield fixed options.

    They will be extended with columns in JS and prior to validation.
    """
    yield 'p', 'Performance for all rows'
    yield 'm', 'Mechanical for all rows'
    yield 's', 'Sync for all rows'
    yield None, '-' * 40


class RoyaltyCalculationForm(forms.Form):
    """The form for royalty calculations.
    """
    class Meta:
        fields = ('in_file',)

    class Media:
        css = {'all': ('admin/css/forms.css',)}
        js = (
            'admin/js/vendor/jquery/jquery.js',
            'admin/js/jquery.init.js')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.file = None

    ALGOS = [
        ('share', 'Split by calculated share.'),
        ('fee', 'Split by manuscript share and apply fees.'),
    ]

    in_file = forms.FileField(
        label='Incoming statement',
        help_text='A CSV file with a header row is required.'
    )
    work_id_column = forms.ChoiceField(
        choices=[], label='Work ID', initial=1,
        help_text='Select the column containing work IDs.'
    )
    work_id_source = forms.ChoiceField(
        choices=get_id_sources, initial=settings.PUBLISHER_CODE,
        label='Work ID Source',
        help_text='You or CWR acknowledgement source.',
    )
    amount_column = forms.ChoiceField(
        choices=[], label='Amount', initial=-1,
        help_text='Select the column containing received amount.'
    )
    right_type_column = forms.ChoiceField(
        choices=get_right_types, initial='p', label='Right type',
        help_text='Select the right type or the column specifying it.'
    )
    algo = forms.ChoiceField(
        choices=ALGOS, initial='fee', label='Algorithm type',
        help_text='Choose the algorithm type, see user manual for details.'
    )
    default_fee = forms.DecimalField(
        max_digits=5, decimal_places=2, initial=0, label='Default fee (%)',
        help_text='Used if no fee is present in the database.'
    )

    def is_valid(self):
        """Append additional choices to various fields, prior to the actual
        validation.
        """
        self.file = TextIOWrapper(self.files.get('in_file'))
        try:
            csv_reader = csv.DictReader(self.file)
            for i, field in enumerate(csv_reader.fieldnames):
                self.fields['work_id_column'].choices.append((str(i), field))
                self.fields['right_type_column'].choices = list(
                    self.fields['right_type_column'].choices)
                self.fields['right_type_column'].choices.append(
                    (str(i), field))
                self.fields['amount_column'].choices.append((str(i), field))
            valid = super().is_valid()
        except Exception:  # to match user stupidity
            return False
        return valid


class RoyaltyCalculation(object):
    """The process of royalty calculation."""

    def __init__(self, form):
        """Initialization with data from thew form and empty attributes.
        """
        self.file = form.file
        for key, value in form.cleaned_data.items():
            setattr(self, key, value)
        self.wc = int(form.cleaned_data.get('work_id_column'))
        self.right = form.cleaned_data.get('right_type_column')
        if self.right in ['p', 'm', 's']:
            self.rc = None
        else:
            self.rc = int(self.right)
            self.right = None
        self.ac = int(form.cleaned_data.get('amount_column'))
        self.writer_ids = set()
        self.writers = {}
        self.works = defaultdict(list)

    @property
    def filename(self):
        """Return the filename of the output file."""
        in_name = self.in_file.name.rsplit('.', 1)[0]
        return in_name + '-output-' + self.algo + '.csv'

    @property
    def fieldnames(self):
        """Return the list of field names in the output file."""
        self.file.seek(0)
        csv_reader = csv.DictReader(self.file)
        fieldnames = csv_reader.fieldnames
        if self.algo == 'share':
            fieldnames.append('Right type')
        fieldnames += [
            'Controlled by publisher (%)',
            'Interested party',
            'Role']
        if self.algo == 'fee':
            fieldnames += [
                'Manuscript share (%)',
                'Share in amount received (%)',
                'Amount before fee',
                'Fee (%)',
                'Fee amount']
        elif self.algo == 'share':
            fieldnames += [
                'Owned share (%)',
                'Share in amount received (%)']
        fieldnames.append('Net amount')
        return fieldnames

    def get_work_ids(self):
        """
        Find work unambiguous work identifiers.

        Returns:
            set of work identifier from the file
        """

        self.file.seek(0)
        csv_reader = csv.reader(self.file)
        work_ids = set()
        for row in csv_reader:
            given_id = row[self.wc]
            if self.work_id_source in ['ISWC', 'ISRC']:
                given_id = given_id.replace('.', '').replace('-', '')
            work_ids.add(given_id)
        return work_ids

    def get_work_queryset(self, work_ids):
        """
        Return the appropriate queryset based on work ID source and ids.

        Returns:
            queryset with :class:`.models.WriterInWork` objects. \
            ``query_id`` has the matched field value.
        """
        qs = WriterInWork.objects.filter(controlled=True)
        if self.work_id_source == settings.PUBLISHER_CODE:
            qs = qs.filter(work___work_id__in=work_ids)
            qs = qs.extra(select={'query_id': "_work_id"})
        elif self.work_id_source == 'ISWC':
            qs = qs.filter(work__iswc__in=work_ids)
            qs = qs.extra(select={'query_id': "iswc"})
        elif self.work_id_source == 'ISRC':
            qs = qs.filter(work__recordings__isrc__in=work_ids)
            qs = qs.extra(select={'query_id': "isrc"})
        else:
            qs = qs.filter(
                work__workacknowledgement__society_code=self.work_id_source)
            qs = qs.filter(
                work__workacknowledgement__remote_work_id__in=work_ids)
            qs = qs.extra(select={
                'query_id':
                    "music_publisher_workacknowledgement.remote_work_id"})
        qs = qs.distinct()
        return qs

    def generate_works_dict(self, qs):
        """Generate the works cache.

        Returns:
            dict (works) of lists (writerinwork) of dicts
        """
        for wiw in qs:
            assert (wiw.work_id is not None)
            self.writer_ids.add(wiw.writer_id)
            d = {
                'writer_id': wiw.writer_id,
                'role': wiw.get_capacity_display(),
                'relative_share': wiw.relative_share,
                'fee': wiw.publisher_fee,
            }
            self.works[wiw.query_id].append(d)

    def generate_writer_dict(self):
        """Generate the writers cache.
        Returns:
            dict (writer) of dicts
        """
        qs = Writer.objects.filter(id__in=self.writer_ids)
        for writer in qs:
            if writer.first_name:
                name = '{}, {} [{}]'.format(
                    writer.last_name,
                    writer.first_name,
                    writer.ipi_name or '')
            else:
                name = '{} [{}]'.format(
                    writer.last_name,
                    writer.ipi_name or '')

            self.writers[writer.id] = {
                'name': name,
                'fee': writer.publisher_fee
            }

    def get_works_and_writers(self):
        """Get work and writer data.

        Extract all work IDs, then perform the queries and put them in
        dictionaries. When the actual file processing happens, no further
        queries are required.
        """

        # the first pass of processing
        work_ids = self.get_work_ids()

        # the first query
        qs = self.get_work_queryset(work_ids)

        self.generate_works_dict(qs)

        # this includes the second query
        self.generate_writer_dict()

    def process_row(self, row):
        """Process one incoming row, yield multiple output rows."""
        # get the identifier and clean
        given_id = row[self.wc]
        if self.work_id_source in ['ISWC', 'ISRC']:
            given_id = given_id.replace('.', '').replace('-', '')

        # get the work, if not found yield error
        work = self.works.get(given_id)

        amount = Decimal(row[self.ac])
        right = (self.right or row[self.rc][0]).lower()
        share_split = {
            'p': settings.PUBLISHING_AGREEMENT_PUBLISHER_PR,
            'm': settings.PUBLISHING_AGREEMENT_PUBLISHER_MR,
            's': settings.PUBLISHING_AGREEMENT_PUBLISHER_SR}[right]

        # Add data to all output rows
        if self.algo == 'share':
            row.append({'p': 'Perf.', 'm': 'Mech.', 's': 'Sync'}[right])
        if not work:
            row.append('')
            row.append('ERROR: Work not found')
            yield row
            return

        controlled = sum([line['relative_share'] for line in work]) / 100
        row.append('{0:.4f}'.format(controlled))

        # Prepare output lines, one per controlled writer in work
        for line in work:

            # Common fields for all algorithms
            out_row = row.copy()
            writer = self.writers[line.get('writer_id')]
            out_row.append(writer['name'])
            out_row.append(line['role'])
            relative_share = line['relative_share'] / 100

            if self.algo == 'fee':
                out_row.append('{0:.4f}'.format(relative_share))
                share = (relative_share / controlled).quantize(
                    Decimal('.000001'))
                amount_before_fee = amount * share
                out_row.append('{0:.6f}'.format(share))
                out_row.append('{}'.format(amount_before_fee))
                fee = (line['fee'] or writer['fee'] or self.default_fee) / 100
                out_row.append('{}'.format(fee))
                fee_amount = amount_before_fee * fee
                out_row.append('{}'.format(fee_amount or '0'))
                net_amount = amount_before_fee - fee_amount
                out_row.append('{}'.format(net_amount))

            elif self.algo == 'share':

                # do not show lines when writers get nothing
                if share_split == Decimal(1):
                    continue

                owned_share = relative_share * (1 - share_split)
                out_row.append('{0:.6f}'.format(owned_share))
                share = (owned_share / controlled).quantize(Decimal('.000001'))
                net_amount = amount * share
                out_row.append('{0:.6f}'.format(share))
                out_row.append('{}'.format(net_amount))

            yield out_row

        else:

            # "Share" algorithm has one additional row with the publisher
            if self.algo == 'share':
                out_row = row.copy()
                out_row.append('{}, [{}]'.format(
                    settings.PUBLISHER_NAME,
                    settings.PUBLISHER_IPI_NAME))
                out_row.append('Original Publisher')
                out_row.append('{0:.6f}'.format(share_split * controlled))
                out_row.append('{0:.6f}'.format(share_split))
                net_amount = amount * share_split
                out_row.append('{}'.format(net_amount))

                yield out_row

    @property
    def out_file_path(self):
        """This method creates the output file and outputs the temporary path.

        Note that the process happens is several passes.
        """
        self.get_works_and_writers()

        # the second pass of processing
        self.file.seek(0)
        csv_reader = csv.reader(self.file)

        f = NamedTemporaryFile(mode='w+', delete=False, encoding='utf8')
        csv_writer = csv.writer(f)
        csv_writer.writerow(self.fieldnames)
        for row in csv_reader:
            for out_row in self.process_row(row):
                csv_writer.writerow(out_row)

        f.filename = self.filename
        f.close()
        return f.name


class RoyaltyCalculationView(PermissionRequiredMixin, FormView):
    """The view for royalty calculations."""

    template_name = 'music_publisher/royalty_calculation.html'
    form_class = RoyaltyCalculationForm
    permission_required = ('music_publisher.can_process_royalties',)

    def render_to_response(self, context, **response_kwargs):
        """Prepare the context, required since we use admin template."""
        context['site_header'] = settings.PUBLISHER_NAME
        context['opts'] = {
            'app_label': 'music_publisher',
            'model_name': 'royaltycalculations',
        }
        context['title'] = 'Royalty Calculations'
        context['has_permission'] = True
        context['is_nav_sidebar_enabled'] = False  # Permission issue
        return super().render_to_response(context, **response_kwargs)

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        """Royalty processing works only with TemporaryFileUploadHandler."""
        request.upload_handlers = [TemporaryFileUploadHandler()]
        return super().dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        """This is where the magic happens."""
        rc = RoyaltyCalculation(form)
        path = rc.out_file_path
        f = open(path, 'rb')
        try:
            return FileResponse(f, filename=rc.filename, as_attachment=False)
        finally:
            os.remove(path)



"""Django templates for CWR generation.

Attributes:
    TEMPLATES_21 (dict): Record templates for CWR 2.1
    TEMPLATES_22 (dict): Record templates for CWR 2.2, based on 2.1
    TEMPLATES_30 (dict): Record templates for CWR 3.0
    TEMPLATES_31 (dict): Record templates for CWR 3.1, based on 3.0
"""

from django.template import Template

TEMPLATES_21 = {
    'HDR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'HDRPB{{ ipi_name_number|rjust:11|slice:"2:" }}'
        '{{ name|ljust:45 }}01.10{{ creation_date|date:"Ymd" }}'
        '{{ creation_date|date:"His" }}{{ creation_date|date:"Ymd" }}'
        '               \r\n{% endautoescape %}'),
    # CWR 2.1 revision 8 "hack" - no sender type field, 11 digit IPI name
    'HDR_8': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'HDR{{ ipi_name_number|rjust:11 }}'
        '{{ name|ljust:45 }}01.10{{ '
        'creation_date|date:"Ymd" }}'
        '{{ creation_date|date:"His" }}{{ '
        'creation_date|date:"Ymd" }}'
        '               \r\n{% endautoescape %}'),
    'GRH': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'GRH{{ transaction_type|ljust:3 }}0000102.10'
        '0000000000  \r\n{% endautoescape %}'),
    'WRK': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        '{{ record_type }}'
        '{{ transaction_sequence|rjust:8 }}00000000'
        '{{ work_title|ljust:60 }}  {{ code|ljust:14 }}'
        '{{ iswc|ljust:11 }}00000000            UNC'
        '{{ duration|date:"His"|default:"000000" }}{{ recorded_indicator }}'
        '      {{ version_type }}  ' + ' ' * 40 + 'N00000000000' +
        ' ' * 51 + 'N'
                   '\r\n{% endautoescape %}'),

    'SPU': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'SPU{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ chain_sequence|rjust:2 }}'
        '{{ code|ljust:9 }}'
        '{{ name|ljust:45 }}'
        ' {{ role|default:"E "|ljust:2}}'
        '000000000{{ ipi_name_number|rjust:11 }}              '
        '{{ pr_society|soc }}{{ pr_share|default:0|cwrshare }}'
        '{{ mr_society|soc }}{{ mr_share|default:0|cwrshare }}'
        '{{ sr_society|soc }}{{ sr_share|default:0|cwrshare }}'
        ' N {{ ipi_base_number|ljust:13 }}'
        '              {{ saan|ljust:14 }}  '
        '{{usa_license|ljust:1}}'
        '\r\n{% endautoescape %}'),
    'SPT': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'SPT{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ code|ljust:9 }}'
        '      {{ pr_share|default:0|cwrshare }}'
        '{{ mr_share|default:0|cwrshare }}'
        '{{ sr_share|default:0|cwrshare }}'
        'I2136N001\r\n{% endautoescape %}'),
    'SWR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'SWR{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ code|ljust:9 }}'
        '{{ last_name|ljust:45 }}{{ first_name|ljust:30 }} '
        '{{ writer_role|ljust:2 }}000000000{{ ipi_name_number|rjust:11 }}'
        '{{ pr_society|soc }}{{ pr_share|default:0|cwrshare }}'
        '{{ mr_society|soc }}{{ mr_share|default:0|cwrshare }}'
        '{{ sr_society|soc }}{{ sr_share|default:0|cwrshare }}'
        ' N  {{ ipi_base_number|ljust:13 }}             \r\n'
        '{% endautoescape %}'),
    'SWT': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'SWT{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ code|ljust:9 }}'
        '{{ pr_share|default:0|cwrshare }}'
        '{{ mr_share|default:0|cwrshare }}'
        '{{ sr_share|default:0|cwrshare }}I2136N001\r\n{% endautoescape %}'),
    'PWR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'PWR{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ publisher_code|ljust:9 }}'
        '{{ publisher_name|ljust:45 }}              '
        '{{ saan|ljust:14 }}'
        '{{ code|ljust:9 }}\r\n{% endautoescape %}'),
    'OPU': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'OPU{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ sequence|rjust:2 }}' +
        ' ' * 54 +
        'YE 00000000000000000000              '
        '   {{ pr_share|default:0|cwrshare }}'
        '   {{ mr_share|default:0|cwrshare }}'
        '   {{ sr_share|default:0|cwrshare }}'
        ' N                                             '
        '\r\n{% endautoescape %}'),
    'OWR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'OWR{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ code|ljust:9 }}'
        '{{ last_name|ljust:45 }}{{ first_name|ljust:30 }}'
        '{{ writer_unknown_indicator|default:" "}}'
        '{{ writer_role|ljust:2 }}000000000{{ ipi_name_number|rjust:11 }}'
        '{{ pr_society|soc }}{{ pr_share|default:0|cwrshare }}'
        '{{ mr_society|soc }}{{ mr_share|default:0|cwrshare }}'
        '{{ sr_society|soc }}{{ sr_share|default:0|cwrshare }}    '
        '{{ ipi_base_number|ljust:13 }}             \r\n{% endautoescape %}'),
    'ALT': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'ALT{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ alternate_title|ljust:60 }}AT  '
        '\r\n{% endautoescape %}'),
    'OWK': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'VER{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ work_title|ljust:60 }}' +
        ' ' * (11 + 2 + 45 + 30 + 60 + 11 + 13 + 45 + 30 + 11 + 13 + 14) +
        '\r\n{% endautoescape %}'),
    'PER': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'PER{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ last_name|ljust:45 }}'
        '{{ first_name|ljust:30 }}                        \r\n'
        '{% endautoescape %}'),
    'REC': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'REC{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}'
        '{{ release_date|default:"00000000" }}' +
        ' ' * 60 + '{{ duration|rjust:6|default:"000000" }}     ' +
        ' ' * 151 +
        '{{ isrc|ljust:12 }}     \r\n{% endautoescape %}'),
    'ORN': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'ORN{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}LIB' + ' ' * 60 +
        '{{ cd_identifier|ljust:15 }}0000{{ library|ljust:60 }}' +
        ' ' * (26 + 12 + 60 + 20) + '0000                  \r\n'
                                    '{% endautoescape %}'),
    'GRT': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'GRT00001{{ transaction_count|rjust:8 }}'
        '{{ record_count|rjust:8 }}   0000000000\r\n{% endautoescape %}'),
    'TRL': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'TRL00001{{ transaction_count|rjust:8 }}'
        '{{ record_count|rjust:8 }}{% endautoescape %}'),
    'OPT': Template(''),
    'OWT': Template(''),
    'XRF': Template(''),
    'MAN': Template('')
}

TEMPLATES_22 = TEMPLATES_21.copy()
TEMPLATES_22.update({
    'HDR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'HDRPB{{ ipi_name_number|rjust:11|slice:"2:" }}'
        '{{ name|ljust:45 }}01.10{{ creation_date|date:"Ymd" }}'
        '{{ creation_date|date:"His" }}{{ creation_date|date:"Ymd" }}'
        '               2.2002{{ settings.SOFTWARE|ljust:30 }}'
        '{{ settings.SOFTWARE_VERSION|ljust:30 }}\r\n{% endautoescape %}'),
    'HDR_8': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'HDR{{ ipi_name_number|rjust:11 }}'
        '{{ name|ljust:45 }}01.10{{ '
        'creation_date|date:"Ymd" }}'
        '{{ creation_date|date:"His" }}{{ '
        'creation_date|date:"Ymd" }}'
        '               2.2002{{ settings.SOFTWARE|ljust:30 }}'
        '{{ settings.SOFTWARE_VERSION|ljust:30 }}\r\n{% endautoescape %}'),
    'GRH': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'GRH{{ transaction_type|ljust:3 }}0000102.20'
        '0000000000  \r\n{% endautoescape %}'),
    'PWR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'PWR{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ publisher_code|ljust:9 }}'
        '{{ publisher_name|ljust:45 }}              '
        '{{ saan|ljust:14 }}'
        '{{ code|ljust:9 }}01\r\n{% endautoescape %}'),
    'ORN': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'ORN{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}LIB' + ' ' * 60 +
        '{{ cd_identifier|ljust:15 }}0000{{ library|ljust:60 }}' +
        ' ' * (26 + 12 + 60 + 20) + '0000' + ' ' * (18 + 26 + 42) + '\r\n'
        '{% endautoescape %}'),
    'REC': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'REC{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}'
        '{{ release_date|default:"00000000" }}' +
        ' ' * 60 + '{{ duration|rjust:6|default:"000000" }}     ' +
        ' ' * 151 +
        '{{ isrc|ljust:12 }}     '
        '{{ recording_title|ljust:60 }}{{ version_title|ljust:60 }}'
        '{{ display_artist|ljust:60 }}{{ record_label.name|ljust:60 }}'
        '{{ isrc_validity|ljust:20 }}{{ code|ljust:14 }}'
        '\r\n{% endautoescape %}'),
    'XRF': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'XRF{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ organization.code|soc }}'
        '{{ identifier|ljust:14 }}WY\r\n{% endautoescape %}'),
})

TEMPLATES_30 = {
    'HDR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'HDRPB{{ code|ljust:4 }}'
        '{{ name|ljust:45 }}' + ' ' * 11 +
        '{{ creation_date|date:"Ymd" }}'
        '{{ creation_date|date:"His" }}{{ creation_date|date:"Ymd" }}' +
        ' ' * 15 + '3.0000{{ settings.SOFTWARE|ljust:30 }}'
        '{{ settings.SOFTWARE_VERSION|ljust:30 }}'
        '{{ filename|ljust:27 }}\r\n{% endautoescape %}'),
    'GRH': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'GRH{{ transaction_type|ljust:3 }}0000103.000000000000'
        '\r\n{% endautoescape %}'),
    'WRK': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'WRK{{ transaction_sequence|rjust:8 }}00000000'
        '{{ work_title|ljust:60 }}  {{ code|ljust:14 }}'
        '{{ iswc|ljust:11 }}00000000            UNC'
        '{{ duration|date:"His"|default:"000000" }}{{ recorded_indicator }}'
        '      {{ version_type }}N00000000000' + ' ' * 51 +
        'N\r\n{% endautoescape %}'),
    'SPU': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'SPU{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}01'
        '{{ code|ljust:9 }}'
        '{{ name|ljust:45 }}N{{ role|default:"E "|ljust:2}}'
        '000000000{{ ipi_name_number|rjust:11 }}'
        '{{ ipi_base_number|ljust:13 }} \r\n{% endautoescape %}'),
    'SPT': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'SPT{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}001{{ code|ljust:9 }}'
        '{{ pr_share|default:0|cwrshare }}'
        '{{ mr_share|default:0|cwrshare }}'
        '{{ sr_share|default:0|cwrshare }}I2136'
        '{{ pr_society|ljust:4 }}'
        '{{ mr_society|ljust:4 }}'
        '{{ sr_society|ljust:4 }}'
        + ' ' * 32 + '0000\r\n{% endautoescape %}'),
    'SWR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'SWR{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ code|ljust:9 }}'
        '{{ last_name|ljust:45 }}{{ first_name|ljust:30 }}N'
        '{{ writer_role|ljust:2 }}000000000{{ ipi_name_number|rjust:11 }}'
        '{{ ipi_base_number|ljust:13 }} N  \r\n'
        '{% endautoescape %}'),
    'SWT': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'SWT{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}001{{ code|ljust:9 }}'
        '{{ pr_share|default:0|cwrshare }}'
        '{{ mr_share|default:0|cwrshare }}'
        '{{ sr_share|default:0|cwrshare }}I2136{{ pr_society|ljust:4 }}'
        '{{ mr_society|ljust:4 }}{{ sr_society|ljust:4 }}'
        + ' ' * 32 + '0000\r\n{% endautoescape %}'),
    'OWT': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'OWT{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}001{{ code|ljust:9 }}'
        '{{ pr_share|default:0|cwrshare }}'
        '{{ mr_share|default:0|cwrshare }}'
        '{{ pr_share|default:0|cwrshare }}'
        'I2136{{ pr_society|ljust:4 }}'
        '{{ mr_society|ljust:4 }}{{ sr_society|ljust:4 }}'
        + ' ' * 32 + '0000\r\n{% endautoescape %}'),
    'PWR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'PWR{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}'
        '{{ publisher_sequence|rjust:2 }}{{ publisher_code|ljust:9 }}'
        '{{ code|ljust:9 }}' + ' ' * 14 +
        '{{ publisher_pr_society|ljust:4 }}'
        '{{ saan|ljust:14 }}'
        '{{ original_publishers.0.agreement.agreement_type.code|ljust:2 }}'
        '\r\n{% endautoescape %}'),
    'OPU': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'OPU{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ sequence|rjust:2 }}' +
        ' ' * 54 +
        'YE 00000000000000000000              \r\n{% endautoescape %}'),
    'OPT': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'OPT{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}001         '
        '{{ pr_share|default:0|cwrshare }}'
        '{{ mr_share|default:0|cwrshare }}'
        '{{ sr_share|default:0|cwrshare }}I2136' + ' ' * 44 + '0000\r\n'
        '{% endautoescape %}'),
    'OWR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'OWR{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ code|ljust:9 }}'
        '{{ last_name|ljust:45 }}{{ first_name|ljust:30 }}'
        '{{ writer_unknown_indicator|default:"N"}}'
        '{{ writer_role|ljust:2 }}000000000{{ ipi_name_number|rjust:11 }}'
        '{{ ipi_base_number|ljust:13 }} N  \r\n'
        '{% endautoescape %}'),
    'ALT': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'ALT{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ alternate_title|ljust:60 }}AT  '
        '\r\n{% endautoescape %}'),
    'OWK': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'OWK{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ work_title|ljust:60 }}' +
        ' ' * (11 + 14 + 50 + 8 + 45 + 30 + 11 + 13 + 45 + 30 + 11 + 13) +
        '\r\n{% endautoescape %}'),
    'PER': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'PER{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ last_name|ljust:45 }}'
        '{{ first_name|ljust:30 }}' + ' ' * 11 +
        '{{ isni|ljust:16 }}     \r\n{% endautoescape %}'),
    'REC': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'REC{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}'
        '{{ release_date|default:"00000000" }}'
        '{{ duration|rjust:6|default:"000000" }}'
        '{{ isrc|ljust:12 }}{{ recording_title|ljust:60 }}'
        '{{ version_title|ljust:60 }}{{ display_artist|ljust:60 }}' +
        ' ' * 11 +
        '{{ recording_artist.isni|ljust:16 }}{{ record_label.name|ljust:60 }}'
        '{{ isrc_validity|ljust:20 }}{{ code|ljust:14 }}\r\n'
        '{% endautoescape %}'),
    'ORN': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'ORN{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}LIB'
        '{{ cd_identifier|ljust:15 }}0000'
        '{{ library|ljust:60 }}' +
        ' ' * (60 + 60 + 1 + 12 + 60 + 20) + '0000' +
        ' ' * (19 + 26 + 21 + 40) + '\r\n{% endautoescape %}'),
    'XRF': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'XRF{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ organization.code|ljust:4 }}'
        '{{ identifier|ljust:14 }}WY\r\n{% endautoescape %}'),
    'ISR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'ISR{{ transaction_sequence|rjust:8 }}00000000'
        '{{ work_title|ljust:60 }}  {{ work_id|ljust:14 }}'
        '{{ iswc|ljust:11 }}{{ indicator|ljust:1 }}\r\n{% endautoescape %}'),
    'WRI': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'WRI{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ code|ljust:9 }}'
        '{{ ipi_name_number|rjust:11 }}{{ last_name|ljust:45 }}'
        '{{ first_name|ljust:30 }}{{ capacity|ljust:2 }}\r\n'
        '{% endautoescape %}'),
    'GRT': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'GRT00001{{ transaction_count|rjust:8 }}'
        '{{ record_count|rjust:8 }}\r\n{% endautoescape %}'),
    'TRL': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'TRL00001{{ transaction_count|rjust:8 }}'
        '{{ record_count|rjust:8 }}{% endautoescape %}'),
    'MAN': Template('')
}

TEMPLATES_31 = TEMPLATES_30.copy()
TEMPLATES_31.update({
    'HDR': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'HDRPB{{ code|ljust:4 }}'
        '{{ name|ljust:45 }}' + ' ' * 11 +
        '{{ creation_date|date:"Ymd" }}'
        '{{ creation_date|date:"His" }}{{ creation_date|date:"Ymd" }}' +
        ' ' * 15 + '3.1000{{ settings.SOFTWARE|ljust:30 }}'
        '{{ settings.SOFTWARE_VERSION|ljust:30 }}'
        '{{ filename|ljust:27 }}\r\n{% endautoescape %}'),
    'GRH': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'GRHWRK0000103.100000000000'
        '\r\n{% endautoescape %}'),
    'MAN': Template(
        '{% load cwr_generators %}{% autoescape off %}'
        'MAN{{ transaction_sequence|rjust:8 }}'
        '{{ record_sequence|rjust:8 }}{{ code|ljust:9 }}'
        '{{ share|cwrshare }}{{ share|cwrshare }}'
        '{{ share|cwrshare }}\r\n{% endautoescape %}')})



"""Create society tuple and dict.

    Attributes:
        SOCIETIES (tuple): (tis-n, Name (Country))
        SOCIETY_DICT (dict): {tis-n, Name (Country)}
"""

import csv
import os
from collections import OrderedDict

dir_path = os.path.dirname(os.path.realpath(__file__))
path = os.path.join(dir_path, 'societies.csv')

with open(path, 'r') as f:
    reader = csv.reader(f)
    SOCIETIES = sorted(
        ((str(row[0]), '{} ({})'.format(row[1], row[2]))
         for row in reader),
        key=lambda row: row[1])

SOCIETY_DICT = OrderedDict(SOCIETIES)


"""Django app definition for :mod:`music_publisher`."""

from django.apps import AppConfig
from .validators import validate_settings


class MusicPublisherConfig(AppConfig):
    """Configuration for Music Publisher app.

    Attributes:
        label (str): app label
        name (str): app name
        verbose_name (str): app verbose name
    """

    name = 'music_publisher'
    label = 'music_publisher'
    verbose_name = 'Music Publisher'

    def ready(self):
        """Validate settings when ready to prevent deployments with invalid
        settings. """
        validate_settings()



from .models import (
    Writer, Work, Recording, Artist, Release, CommercialRelease, 
    LibraryRelease, Label, Track, Playlist)
from rest_framework import viewsets, serializers, permissions, renderers
from rest_framework.response import Response
from django.utils.timezone import now
from django.http import Http404


class DjangoModelPermissionsIncludingView(permissions.DjangoModelPermissions):
    """Requires the user to have proper permissions, including view."""
    perms_map = {
        'GET': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': ['%(app_label)s.view_%(model_name)s'],
        'HEAD': ['%(app_label)s.view_%(model_name)s'],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
        'DELETE': ['%(app_label)s.delete_%(model_name)s'],
    }


class ArtistNestedSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Artist
        fields = ['name', 'image', 'description', 'url']

    name = serializers.CharField(source='__str__', read_only=True)


class LabelNestedSerializer(serializers.ModelSerializer):
    class Meta:
        model = Label
        fields = ['name', 'image', 'description']


class WriterNestedSerializer(serializers.ModelSerializer):
    class Meta:
        model = Writer
        fields = ['last_name', 'first_name', 'ipi_name', 'image', 'description']


class WriterNamesField(serializers.RelatedField):
    def to_representation(self, value):
        for writer in value.distinct():
            yield WriterNestedSerializer(writer).data


class WorkNestedSerializer(serializers.ModelSerializer):
    class Meta:
        model = Work
        fields = ['title', 'work_id', 'iswc', 'writers']

    writers = WriterNamesField(read_only=True)


class ReleaseListSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Release
        fields = [
            'title', 'image', 'description', 'release_label', 'release_date',
            'ean', 'url']
    title = serializers.CharField(source='release_title', read_only=True)
    release_label = LabelNestedSerializer(read_only=True)


class RecordingInTrackSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Recording
        fields = [
            'title', 
            'recording_id', 'isrc',
            'duration', 'release_date', 'audio_file',
            'artist',
            'record_label',
            'work',
        ]

    artist = ArtistNestedSerializer(read_only=True)
    work = WorkNestedSerializer(read_only=True)
    record_label = LabelNestedSerializer(read_only=True)


class TrackNestedSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Track
        fields = ['cut_number', 'recording']

    recording = RecordingInTrackSerializer(read_only=True)


class ReleaseDetailSerializer(serializers.ModelSerializer):
    class Meta:
        model = Release
        fields = [
            'title', 'image', 'description', 'release_label', 'release_date',
            'ean', 'tracks']
    title = serializers.CharField(source='release_title', read_only=True)
    release_label = LabelNestedSerializer(read_only=True)
    tracks = TrackNestedSerializer(many=True, read_only=True)


class ReleaseViewSet(viewsets.ReadOnlyModelViewSet):
    """
    API endpoint for releases with images or descriptions.

    Requires authenticated user with appropriate ``view`` permission.

    Note that all related information, including files are included in
    this view.

    """
    queryset = Release.objects \
        .exclude(library__isnull=True, cd_identifier__isnull=False) \
        .exclude(image='', description='')\
        .select_related('release_label')

    permission_classes = [DjangoModelPermissionsIncludingView]

    def get_serializer(self, *args, **kwargs):
        if kwargs.get('many'):
            serializer_class = ReleaseListSerializer
        else:
            serializer_class = ReleaseDetailSerializer
        kwargs.setdefault('context', self.get_serializer_context())
        return serializer_class(*args, **kwargs)


class RecordingInArtistSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Recording
        fields = [
            'title', 'isrc', 'duration', 'release_date', 'audio_file',
            'record_label',
            'work',
        ]

    work = WorkNestedSerializer(read_only=True)
    record_label = LabelNestedSerializer(read_only=True)


class ArtistDetailSerializer(serializers.ModelSerializer):
    class Meta:

        model = Release
        fields = ['name', 'image', 'description', 'recordings']

    name = serializers.CharField(source='__str__', read_only=True)
    recordings = RecordingInArtistSerializer(read_only=True, many=True)


class ArtistViewSet(viewsets.ReadOnlyModelViewSet):
    """
    API endpoint for artists with images or descriptions.

    Requires authenticated user with appropriate ``view`` permission.

    Note that all related information, including files are included in
    this view.
    """
    queryset = Artist.objects\
        .exclude(image='', description='')

    permission_classes = [DjangoModelPermissionsIncludingView]

    def get_serializer(self, *args, **kwargs):
        if kwargs.get('many'):
            serializer_class = ArtistNestedSerializer
        else:
            serializer_class = ArtistDetailSerializer
        kwargs.setdefault('context', self.get_serializer_context())
        return serializer_class(*args, **kwargs)


class PlaylistSerializer(serializers.ModelSerializer):
    class Meta:

        model = Playlist
        fields = ['release_title', 'image', 'description', 'recordings']
    recordings = RecordingInTrackSerializer(many=True, read_only=True)


class PlaylistViewSet(viewsets.ViewSet):
    """Endpoint for retrieval of playlists, using secret URL.
    You can find the URL in the playlist ``change view``.
    """
    queryset = Playlist.objects.none()
    lookup_field = 'cd_identifier'
    serializer_class = PlaylistSerializer

    def retrieve(self, request, cd_identifier=None):
        playlist = Playlist.objects.filter(cd_identifier=cd_identifier)
        playlist = playlist.exclude(release_date__lt=now())
        playlist = playlist.first()
        if playlist is None:
            raise Http404
        return Response(PlaylistSerializer(
            playlist, context={'request': request}).data)


class IsSuperuser(permissions.BasePermission):
    """
    Allows access only to authenticated superusers.
    """

    def has_permission(self, request, view):
        return bool(
            request.user and request.user.is_authenticated and 
            request.user.is_superuser)


class BackupViewSet(viewsets.ViewSet):
    """Creates a large json file with all works and all related metadata."""

    permission_classes = [IsSuperuser]
    renderer_classes = [renderers.JSONRenderer]

    def list(self, request, *args, **kwargs):
        d = {}
        d.update(Work.objects.get_dict(Work.objects.all()))
        d.update(CommercialRelease.objects.get_dict(Release.objects.all()))
        return Response(d)



# Generated by Django 3.2.6 on 2021-12-02 22:02

from django.db import migrations, models
import django.db.models.deletion
import music_publisher.base
import music_publisher.validators


class Migration(migrations.Migration):

    dependencies = [
        ('music_publisher', '0003_cwrexport_created_on'),
    ]

    operations = [
        migrations.CreateModel(
            name='Playlist',
            fields=[
            ],
            options={
                'verbose_name': 'Playlist',
                'verbose_name_plural': 'Playlists',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('music_publisher.release',),
        ),
        migrations.AlterModelOptions(
            name='artistinwork',
            options={'ordering': ('artist__last_name', 'artist__first_name'), 'verbose_name': 'Artist performing', 'verbose_name_plural': 'Artists performing (not mentioned in recordings section)'},
        ),
        migrations.AddField(
            model_name='artist',
            name='description',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='artist',
            name='image',
            field=models.ImageField(blank=True, max_length=255, upload_to=music_publisher.base.upload_to),
        ),
        migrations.AddField(
            model_name='label',
            name='description',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='label',
            name='image',
            field=models.ImageField(blank=True, max_length=255, upload_to=music_publisher.base.upload_to, verbose_name='Logo'),
        ),
        migrations.AddField(
            model_name='recording',
            name='audio_file',
            field=models.FileField(blank=True, max_length=255, upload_to=music_publisher.base.upload_to),
        ),
        migrations.AddField(
            model_name='release',
            name='description',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='release',
            name='image',
            field=models.ImageField(blank=True, max_length=255, upload_to=music_publisher.base.upload_to, verbose_name='Cover Art'),
        ),
        migrations.AddField(
            model_name='writer',
            name='description',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='writer',
            name='image',
            field=models.ImageField(blank=True, max_length=255, upload_to=music_publisher.base.upload_to),
        ),
        migrations.AlterField(
            model_name='recording',
            name='artist',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='recordings', to='music_publisher.artist', verbose_name='Recording Artist'),
        ),
        migrations.AlterField(
            model_name='track',
            name='release',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tracks', to='music_publisher.release'),
        ),
        migrations.AlterField(
            model_name='work',
            name='writers',
            field=models.ManyToManyField(related_name='works', through='music_publisher.WriterInWork', to='music_publisher.Writer'),
        ),
        migrations.AlterField(
            model_name='writer',
            name='saan',
            field=models.CharField(blank=True, help_text='Use this field for a general original publishing agreement.', max_length=14, null=True, validators=[music_publisher.validators.CWRFieldValidator('saan')], verbose_name='SAAN'),
        ),
    ]



"""Filters used in generation of CWR files.

"""

from decimal import Decimal, ROUND_HALF_UP

from django import template
from django.conf import settings

from music_publisher import models
from music_metadata.territories.territory import Territory

register = template.Library()

@register.filter(name='rjust')
def rjust(value, length):
    """Format general numeric fields."""

    if value is None or value == '':
        value = '0'
    else:
        value = str(value)
    value = value.rjust(length, '0')[0:length]
    return value


@register.filter(name='ljust')
def ljust(value, length):
    """Format general alphanumeric fields."""

    if value is None:
        value = ''
    else:
        value = str(value)
    value = value.ljust(length, ' ')[0:length]
    return value


@register.filter(name='soc')
def soc(value):
    """Format society fields."""

    if not value:
        return '   '
    value = value.rjust(3, '0')
    return value


@register.filter(name='cwrshare')
def cwrshare(value):
    """Get CWR-compatible output for share fields"""
    value = (value * Decimal('10000')).quantize(
        Decimal('1.'), rounding=ROUND_HALF_UP)
    value = int(value)
    return '{:05d}'.format(value)



"""Contains base (abstract) classes used in :mod:`.models`
"""

import re

import django.core.exceptions
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

from .validators import CWRFieldValidator
from .societies import SOCIETIES
from uuid import uuid4
import base64


def upload_to(instance, filename):
    ext = filename.rsplit('.', 1)[-1]
    folder = instance._meta.model_name
    fn = base64.urlsafe_b64encode(uuid4().bytes).rstrip(b'=')
    return f'{folder}/{fn}.{ext}'


class NotesManager(models.Manager):
    """Manager for objects inheriting from :class:`NotesBase`.

    Defers :attr:`NotesBase.notes` field.
    """
    def get_queryset(self):
        """Defer :attr:`NotesBase.notes` field."""
        qs = super().get_queryset()
        qs = qs.defer('notes')
        return qs


class NotesBase(models.Model):
    """Abstract class for all classes that have notes.

    Attributes:
        notes (django.db.models.TextField): Notes, free internal text field
    """

    class Meta:
        abstract = True

    objects = NotesManager()

    notes = models.TextField(blank=True)


class DescriptionBase(models.Model):
    """Abstract class for all classes that have publicly visible descriptions.

    Attributes:
        description (django.db.models.TextField): Public description
    """

    class Meta:
        abstract = True

    description = models.TextField(blank=True)


class TitleBase(models.Model):
    """Abstract class for all classes that have a title.

    Attributes:
        title (django.db.models.CharField): Title, used in work title,
            alternate title, etc.
    """

    class Meta:
        abstract = True

    title = models.CharField(
        max_length=60, db_index=True,
        validators=(CWRFieldValidator('title'),))

    def __str__(self):
        return self.title


class PersonBase(models.Model):
    """Base class for all classes that contain people with first and last name.

    This includes writers and artists. For bands, only the last name field is
    used.

    Attributes:
        first_name (django.db.models.CharField): First Name
        last_name (django.db.models.CharField): Last Name
    """

    class Meta:
        abstract = True

    first_name = models.CharField(
        max_length=30, blank=True,
        validators=(CWRFieldValidator('name'),))
    last_name = models.CharField(
        max_length=45, db_index=True,
        validators=(CWRFieldValidator('name'),))
    image = models.ImageField(max_length=255, upload_to=upload_to, blank=True)

    def __str__(self):
        if self.first_name:
            return '{0.first_name} {0.last_name}'.format(self).upper()
        return self.last_name.upper()


class SocietyAffiliationBase(models.Model):
    """Abstract base for all objects with CMO affiliations

    Attributes:
        pr_society (django.db.models.CharField):
            Performing Rights Society Code
        mr_society (django.db.models.CharField):
            Mechanical Rights Society Code
        sr_society (django.db.models.CharField):
            Sync. Rights Society Code
    """

    class Meta:
        abstract = True

    pr_society = models.CharField(
        'Performance rights society', max_length=3, blank=True, null=True,
        choices=SOCIETIES + [('99', 'NO SOCIETY')])
    mr_society = models.CharField(
        'Mechanical rights society', max_length=3, blank=True, null=True,
        choices=SOCIETIES)
    sr_society = models.CharField(
        'Synchronization rights society', max_length=3, blank=True, null=True,
        choices=SOCIETIES)


class IPIBase(models.Model):
    """Abstract base for all objects containing IPI numbers.

    Attributes:
        ipi_base (django.db.models.CharField): IPI Base Number
        ipi_name (django.db.models.CharField): IPI Name Number
        _can_be_controlled (django.db.models.BooleanField):
            used to determine if there is enough data for a writer
            to be controlled.
    """

    class Meta:
        abstract = True

    ipi_name = models.CharField(
        'IPI name #', max_length=11, blank=True, null=True, unique=True,
        validators=(CWRFieldValidator('ipi_name'),))
    ipi_base = models.CharField(
        'IPI base #', max_length=15, blank=True, null=True,
        validators=(CWRFieldValidator('ipi_base'),))

    _can_be_controlled = models.BooleanField(
        verbose_name='Can be controlled', editable=False, default=False)

    def clean_fields(self, *args, **kwargs):
        """
        Data cleanup, allowing various import formats to be converted into
        consistently formatted data.
        """
        if self.ipi_name:
            self.ipi_name = self.ipi_name.zfill(11)
        if self.ipi_base:
            self.ipi_base = self.ipi_base.replace('.', '').upper()
            self.ipi_base = re.sub(
                r'(I).?(\d{9}).?(\d)',
                r'\1-\2-\3',
                self.ipi_base)
        return super().clean_fields(*args, **kwargs)


class IPIWithGeneralAgreementBase(IPIBase, SocietyAffiliationBase):
    """Abstract base for all objects with general agreements.

    Attributes:
        saan (django.db.models.CharField):
            Society-assigned agreement number, in this context it is used for
            general agreements, for specific agreements use
            :attr:`.models.WriterInWork.saan`.
        generally_controlled (django.db.models.BooleanField):
            flags if a writer is generally controlled (in all works)
        publisher_fee (django.db.models.DecimalField):
            this field is used in calculating publishing fees
    """

    class Meta:
        abstract = True

    saan = models.CharField(
        'SAAN',
        help_text='Use this field for a general original publishing '
            'agreement.',
        validators=(CWRFieldValidator('saan'),),
        max_length=14, blank=True, null=True)

    generally_controlled = models.BooleanField(
        'General agreement', default=False)
    publisher_fee = models.DecimalField(
        max_digits=5, decimal_places=2, blank=True, null=True,
        validators=[MinValueValidator(0), MaxValueValidator(100)])

    def clean(self):
        """Clean the data and validate."""

        self._can_be_controlled = (
                bool(self.ipi_name) & bool(self.pr_society))
        if self.ipi_name == '00000000000':
            self.ipi_name = None
        if self.pr_society == '99':
            self.pr_society = None

        d = {}
        if not self.generally_controlled:
            if self.saan:
                d['saan'] = 'Only for a general agreement.'
            if self.publisher_fee:
                d['publisher_fee'] = 'Only for a general agreement.'
        else:
            if not self._can_be_controlled:
                d['generally_controlled'] = (
                    'IPI name number and PR society fields are required for '
                    'a controlled writer. See "Writers" in the user manual.')
        if d:
            raise django.core.exceptions.ValidationError(d)

    def clean_fields(self, *args, **kwargs):
        """
        Data cleanup, allowing various import formats to be converted into
        consistently formatted data.
        """
        if self.saan:
            self.saan = self.saan.upper()  # only in CWR, uppercase anyway
        super().clean_fields(*args, **kwargs)


class ArtistBase(PersonBase, NotesBase, DescriptionBase):
    """Performing artist base class.

    Attributes:
        isni (django.db.models.CharField): International Standard Name Id
    """

    class Meta:
        verbose_name = 'Performing Artist'
        verbose_name_plural = 'Performing Artists'
        ordering = ('last_name', 'first_name', '-id')
        abstract = True

    isni = models.CharField(
        'ISNI',
        max_length=16, blank=True, null=True, unique=True,
        validators=(CWRFieldValidator('isni'),))

    def clean_fields(self, *args, **kwargs):
        """ISNI cleanup"""
        if self.isni:
            self.isni = self.isni.rjust(16, '0').upper()
        return models.Model.clean_fields(self, *args, **kwargs)


class WriterBase(
        PersonBase, IPIWithGeneralAgreementBase, NotesBase, DescriptionBase):
    """Base class for writers.
    """

    class Meta:
        ordering = ('last_name', 'first_name', 'ipi_name', '-id')
        verbose_name_plural = 'Writers'
        abstract = True


class LabelBase(NotesBase, DescriptionBase):
    """Music Label base class.

    Attributes:
        name (django.db.models.CharField): Label Name
    """

    class Meta:
        verbose_name_plural = 'Music Labels'
        ordering = ('name',)
        abstract = True

    name = models.CharField(
        max_length=60, unique=True,
        validators=(CWRFieldValidator('name'),))
    image = models.ImageField(
        'Logo', max_length=255, upload_to=upload_to, blank=True)


class LibraryBase(models.Model):
    """Music Library base class.

    Attributes:
        name (django.db.models.CharField): Library Name
    """

    class Meta:
        verbose_name_plural = 'Music Libraries'
        ordering = ('name',)
        abstract = True

    name = models.CharField(
        max_length=60, unique=True,
        validators=(CWRFieldValidator('name'),))


class ReleaseBase(DescriptionBase):
    """Music Release base class

    Attributes:
        cd_identifier (django.db.models.CharField): CD Identifier, used when \
        origin is library
        library (django.db.models.CharField): Library Name
        release_date (django.db.models.DateField): Date of the release
        ean (django.db.models.CharField): EAN code
        release_label (django.db.models.CharField): Label Name
        release_title (django.db.models.CharField): Title of the release
    """

    class Meta:
        ordering = ('release_title', 'cd_identifier', '-id')
        abstract = True

    cd_identifier = models.CharField(
        'CD identifier',
        max_length=15, blank=True, null=True, unique=True,
        validators=(CWRFieldValidator('name'),))
    release_date = models.DateField(
        blank=True, null=True)
    release_title = models.CharField(
        'Release (album) title ',
        max_length=60, blank=True, null=True,
        validators=(CWRFieldValidator('title'),))
    ean = models.CharField(
        'Release (album) EAN',
        max_length=13, blank=True, null=True, unique=True,
        validators=(CWRFieldValidator('ean'),))
    image = models.ImageField(
        'Cover Art', max_length=255, upload_to=upload_to, blank=True)



# Generated by Django 3.2.6 on 2021-08-16 15:15

from django.db import migrations, models
from django.utils.timezone import datetime, make_aware, get_fixed_timezone


def set_created_on(apps, schema_editor):
    cls = apps.get_model('music_publisher', 'CWRExport')
    qs = cls.objects.filter(created_on__isnull=True)
    for obj in qs:
        if obj.nwr_rev in ['NWR', 'NW2', 'REV', 'RE2']:
            s = obj.cwr[64:78]
        else:
            s = obj.cwr[65:79]
        obj.created_on = make_aware(
            datetime.strptime(s, '%Y%m%d%H%M%S'), get_fixed_timezone(0))
        obj.save()


class Migration(migrations.Migration):

    dependencies = [
        ('music_publisher', '0002_mayday'),
    ]

    operations = [
        migrations.AddField(
            model_name='cwrexport',
            name='created_on',
            field=models.DateTimeField(editable=False, null=True),
        ),
        migrations.RunPython(
            code=set_created_on,
            reverse_code=migrations.operations.special.RunPython.noop,
        ),
    ]



"Forms and formsets."

import re
from datetime import datetime
from decimal import Decimal

from django.core.exceptions import ValidationError
from django.forms import (
    BooleanField, FileField, ModelForm, NullBooleanField, Select,
)
from django.forms.models import BaseInlineFormSet
from django.urls import reverse

from .models import ACKImport, Work


class LibraryReleaseForm(ModelForm):
    """Custom form for :class:`.models.LibraryRelease`.
    """

    def __init__(self, *args, **kwargs):
        """Make cd_identifier and library fields required."""
        super().__init__(*args, **kwargs)
        self.fields['cd_identifier'].required = True
        self.fields['library'].required = True


class PlaylistForm(ModelForm):
    """Custom form for :class:`.models.LibraryRelease`.
    """

    def __init__(self, *args, **kwargs):
        """Make cd_identifier and library fields required."""
        super().__init__(*args, **kwargs)
        self.fields['release_title'].required = True
        self.fields['release_title'].label = 'Playlist title'
        self.fields['release_date'].label = 'Valid Until'


class AlternateTitleFormSet(BaseInlineFormSet):
    """Formset for :class:`AlternateTitleInline`.
    """

    orig_cap = ['C ', 'A ', 'CA']

    def clean(self):
        """Performs these checks:
            if suffix is used, then validates the total length

        Returns:
            None

        Raises:
            ValidationError
        """
        super().clean()
        work_title_len = len(self.instance.title)
        for form in self.forms:
            if not form.is_valid():
                return
            if form.cleaned_data and not form.cleaned_data.get('DELETE'):
                if not form.cleaned_data.get('suffix'):
                    continue
                title_suffix_len = len(form.cleaned_data['title'])
                if work_title_len + title_suffix_len > 60 - 1:
                    form.add_error(
                        'title',
                        'Too long for suffix, work title plus suffix must be '
                        '59 characters or less.')


class WorkForm(ModelForm):
    """Custom form for :class:`.models.Work`.

    Calculate values for readonly field version_type."""

    class Meta:
        model = Work
        fields = [
            'title', 'iswc', 'original_title', 'library_release']

    version_type = NullBooleanField(
        widget=Select(
            choices=(
                (None, ''),
                (True, 'Modification'),
                (False, 'Original Work'))),
        disabled=True,
        required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['version_type'].initial = self.instance.is_modification()


class ACKImportForm(ModelForm):
    """Form used for CWR acknowledgement imports.

    Attributes:
        acknowledgement_file (FileField): Field for file upload
    """

    class Meta:
        model = ACKImport
        fields = ('acknowledgement_file', 'import_iswcs')

    acknowledgement_file = FileField()
    import_iswcs = BooleanField(
        label='Import ISWCs if present', required=False, initial=True)

    RE_HDR_21 = re.compile(
        r'^HDR(?:SO|AA)([ \d]{9})(.{45})01\.10(\d{8})\d{6}(\d{8})')

    RE_HDR_30 = re.compile(
        r'^HDR(?:SO|AA)(.{4})(.{45})(\d{8})\d{6}(\d{8}).{15}3\.0000')

    def clean(self):
        """Perform usual clean, then process the file, returning the content
            field as if it was the TextField."""
        super().clean()
        cd = self.cleaned_data
        ack = cd.get('acknowledgement_file')
        filename = ack.name
        if not (len(filename) in [18, 19] and
                filename[-4:].upper() in ['.V21', '.V22']):
            raise ValidationError('Wrong file name format.')
        self.cleaned_data['filename'] = filename
        content = ack.file.read().decode('latin1')
        match = re.match(self.RE_HDR_21, content)
        if not match:
            match = re.match(self.RE_HDR_30, content)
        if not match:
            raise ValidationError('Incorrect CWR header')
        code, name, date1, date2 = match.groups()
        self.cleaned_data['society_code'] = code.strip().lstrip('0')
        self.cleaned_data['society_name'] = name.strip()
        self.cleaned_data['date'] = datetime.strptime(
            max([date1, date2]), '%Y%m%d').date()
        self.cleaned_data['acknowledgement_file'] = content


class WriterInWorkFormSet(BaseInlineFormSet):
    """Formset for :class:`WriterInWorkInline`.
    """

    orig_cap = ['C ', 'A ', 'CA']

    def clean(self):
        """Performs these checks:
            at least one writer must be controlled,
            at least one writer music be Composer or Composer&Lyricist
            sum of relative shares must be ~100%

        Returns:
            None

        Raises:
            ValidationError
        """
        is_modification = self.instance.is_modification()
        super().clean()
        total = 0
        controlled = False
        has_composer = False
        writers = []
        needs_extended_capacity = is_modification
        for form in self.forms:
            if not form.is_valid():
                return
            if form.cleaned_data and not form.cleaned_data.get('DELETE'):
                writer = form.cleaned_data['writer']
                writers.append(writer)
                total += form.cleaned_data['relative_share']
                if form.cleaned_data['controlled']:
                    controlled = True
                if form.cleaned_data['capacity'] in ['C ', 'CA']:
                    has_composer = True
                if (not is_modification and form.cleaned_data['capacity'] and
                        form.cleaned_data['capacity'] not in self.orig_cap):
                    form.add_error(
                        'capacity',
                        'Not allowed in original works.')
                if (is_modification and form.cleaned_data['capacity'] and
                        form.cleaned_data['capacity'] not in self.orig_cap):
                    needs_extended_capacity = False
        if needs_extended_capacity:
            for form in self.forms:
                form.add_error(
                    'capacity',
                    'At least one must be Arranger, Adaptor or Translator.')
            raise ValidationError(
                'In a modified work, '
                'at least one writer must be Arranger, Adaptor or Translator.')
        if not controlled:
            for form in self.forms:
                form.add_error(
                    'controlled', 'At least one writer must be controlled.')
            raise ValidationError('At least one writer must be controlled.')
        if not has_composer:
            for form in self.forms:
                form.add_error(
                    'capacity',
                    'At least one writer must be Composer or '
                    'Composer&Lyricist.')
            raise ValidationError(
                'At least one writer must be Composer or Composer&Lyricist.')
        if not (Decimal(99.98) <= total <= Decimal(100.02)):
            for form in self.forms:
                form.add_error(
                    'relative_share', 'Sum of manuscript shares must be 100%.')
            raise ValidationError('Sum of manuscript shares must be 100%.')
        if is_modification:
            writer_capacities = {}
            for form in self.forms:
                cd = form.cleaned_data
                if cd['controlled']:
                    writer_capacities[cd['writer'].id] = cd['capacity']
            for form in self.forms:
                cd = form.cleaned_data
                if cd['writer'] and not cd['controlled']:
                    controlled_capacity = writer_capacities.get(
                        cd['writer'].id)
                    if (controlled_capacity and
                            cd['capacity'] != controlled_capacity):
                        form.add_error(
                            'capacity',
                            'Must be same as in controlled line for this '
                            'writer.')


class DataImportForm(ModelForm):
    """Form used for data imports.

    Attributes:
        data_file (FileField): Field for file upload
    """

    class Meta:
        model = ACKImport
        fields = ('data_file',)

    data_file = FileField()
    ignore_unknown_columns = BooleanField(required=False, initial=False)

    def clean(self):
        """
        This is the actual import process, if all goes well,
        the report is saved.

        Raises:
            ValidationError
        """
        super().clean()

        from .data_import import DataImporter
        from io import TextIOWrapper
        from django.db import transaction

        cd = self.cleaned_data
        f = cd.get('data_file')
        report = ''
        with transaction.atomic():
            try:
                importer = DataImporter(TextIOWrapper(f), self.user)
                for work in importer.run():
                    url = reverse(
                        'admin:music_publisher_work_change', args=(work.id,))
                    report += '<a href="{}">{}</a> {}<br/>\n'.format(
                        url, work.work_id, work.title)
                if importer.unknown_keys:
                    if cd.get('ignore_unknown_columns'):
                        report += '<br>\nUNKNOWN COLUMN NAMES:<br>\n'
                        report += '<br>\n'.join(
                            [f'- {key}' for key in sorted(
                                importer.unknown_keys)]
                        )
                    else:
                        raise ValidationError(
                            'Unknown columns: ' +
                            ', '.join(importer.unknown_keys))
                report += importer.report
            except Exception as e:  # user garbage, too many possibilities
                raise ValidationError(str(e))
        self.cleaned_data['report'] = report


